/* started off as a interface to g3vga viewer, now a generic frontend */
/* g3vga - simple g3 FAX fiewer -
   Mitch 30th May 1994 (m.dsouza@mrc-apu.cam.ac.uk) */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <limits.h>
#include <malloc.h>
#include <vga.h>
#include <vgagl.h>
#include <fcntl.h>
#include "vgaui.h"
#include "fontutils.h"
#include "miscutils.h"

#define CTRL_F		6
#define CTRL_P		16
#define HOME		49
#define END		52
#define PAGE_UP		53
#define PAGE_DOWN	54
#define ARROW_UP	65
#define ARROW_DOWN	66
#define ARROW_RIGHT	67
#define ARROW_LEFT	68

#define min(x,y) (x<y?x:y)
#define sqr(x) ((x)*(x))
/*
   Mode    1: 320x200, 16 colors (4 bitplanes)
   Mode    2: 640x200, 16 colors (4 bitplanes)
   Mode    3: 640x350, 16 colors (4 bitplanes)
   Mode    4: 640x480, 16 colors (4 bitplanes)
   Mode    5: 320x200, 256 colors (linear)
   Mode    6: 320x240, 256 colors (Mode X)
   Mode    7: 320x400, 256 colors (Mode X)
   Mode    8: 360x480, 256 colors (Mode X)

 */

/* Everyone should have this one at least these days */
#define DEFMODE 4

#define TITLE_WIDTH 19
#define BORDER_WIDTH 7

short int xmax,
  ymax;
short int bw = BORDER_WIDTH;
static char *background;
static char force,
 *curfile;
#define force_vga_display(x,y) force=1;vga_display(x,y)
#define beep() printf("\007")
#define sbwidth 8
static struct sb {
  char empty[sbwidth];
  char fill[sbwidth];
} sb;
static vga_modeinfo *modeinfo;
static short vp_x,
  vp_y,
  vp_wid,
  vp_ht;

vga_image image = {
	NULL,	/* The image data */
	0,	/* The width of the image */
	0,	/* The height of the image */
	NULL,	/* The rdb colormap */
	0.0,	/* The scale of the image (0-1.0)*/
	NULL	/* The filename of the image */
};
char *progname="vgagui";
char *version="0.3";
char (*vga_key_callback)(char)=NULL;

void
vga_display (short int x, short int y)
{
  int i,
    j,
    y1,
    y2;
  static short int old_x = -1,
    old_y = -1;
  short int sb_start,
    sb_len;

  if ((x < 0 || y < 0 || x > image.wid || y > image.ht) ||
      (!force && x == old_x && y == old_y)) {
    beep ();
    return;
  }
  force = 0;
  sb_start = vp_y + (y * vp_ht) / image.ht;
  sb_len = TITLE_WIDTH + sqr (vp_ht) / image.ht;
  if (image.rgb)
	vga_setpalvec(0,256,image.rgb);
  y1 = y;

  for (j = vp_y, i = y1; j < vp_y + vp_ht && i < image.ht; i++, j++) {

    memset (background, PAGE_COLOR, xmax);
    if (image.data[i] && x < image.wid)
      bcopy (image.data[i] + x, background + sbwidth,
	     min (image.wid - x, vp_wid - sbwidth));
    bcopy (j >= sb_start && j <= sb_start + sb_len +vp_y ? sb.fill : sb.empty,
	   background, sbwidth);
    vga_drawscansegment (background, vp_x, j, vp_wid - 1);
  }

  memset (background, PAGE_COLOR, xmax);
  bcopy (sb.fill,background, sbwidth);
  for (i=j; i<vp_y+vp_ht; i++)
    vga_drawscansegment (background, vp_x, i, vp_wid - 1);

  old_x = x;
  old_y = y;
}

void
update_title (font_t * font, int xpos)
{
  char title[1024],
   *p,
   *q;
  int i;

  p = (p = strrchr (progname, '/')) ? ++p : progname;
  q = (q = strrchr (curfile, '/')) ? ++q : curfile;

  vga_setegacolor (TITLE_BG_COLOR);
  for (i = bw+1; i < bw + TITLE_WIDTH-1; i++) {
    vga_drawline (bw, i, bw + vp_wid, i);
  }

  sprintf (title, "%s (%s) - %s", p, version, q);
  vga_write (title, xpos, bw + 2,
	     font, TITLE_FG_COLOR, TITLE_BG_COLOR, ALIGN_CENTER);
}

void
display (char **fname, int optno)
{
  int i,
    j;
  char *m,
    key;
  short int vga_x = 0, vga_y = 0;
  int arrow_step,
    page_step;
  font_t *tfont,
   *pfont,
   *ifont;
  int MODE;
  int curopt = optno;
  char *newname = NULL;
  int rgbsave[256*3];

  MODE = vga_getdefaultmode ();
  if (MODE == -1)
    MODE = DEFMODE;

  memset (sb.empty, BLACK, sbwidth);
  memset (sb.fill, LIGHTGRAY, sbwidth);
  sb.empty[sbwidth - 1] = sb.fill[sbwidth - 1] = DARKGRAY;

  vga_setmode (MODE);
  vga_getpalvec(0,256,rgbsave);
  modeinfo = vga_getmodeinfo (MODE);

  curfile = image.filename;
  xmax = vga_getxdim () - 1;
  ymax = vga_getydim () - 1;

  /* Viewport dimensions */
  vp_x = bw + 1;
  vp_y = bw + 1 + TITLE_WIDTH;
  vp_wid = xmax - (bw << 1);
  vp_ht = ymax - (bw << 1) - TITLE_WIDTH - 1;

#ifdef DEBUG
  fprintf (stderr, "vp_x = %d, vp_y = %d, vp_wid = %d, vp_ht = %d\n", vp_x, vp_y, vp_wid, vp_ht);
#endif
  /* Page keys scroll by this much */
  page_step = vp_ht;

  /* Arrow keys scroll by this much - i.e. half a page step */
  arrow_step = vp_ht >> 1;

#ifdef DEBUG
  fprintf (stderr, "arrow_step = %d, xmax = %d, ymax = %d\n", arrow_step, xmax, ymax);
#endif
  /* This is used for backgrounding unused fax lines to the page color */
  background = malloc (xmax);

  tfont = malloc (sizeof (font_t));		/* The Title font */
  vga_initfont (TITLE_FONT, tfont, 1, 1);	/* See the Makefile for info */

  pfont = malloc (sizeof (font_t));		/* The Prompt Font */
  vga_initfont (PROMPT_FONT, pfont, 1, 1);	/* See the Makefile for info */

  ifont = malloc (sizeof (font_t));		/* The usr Input Font */
  vga_initfont (INPUT_FONT, ifont, 1, 1);	/* See the Makefile for info */

  vga_setcolor (BLACK);
  vga_drawline (bw, i, bw + vp_wid, i);
  vga_drawline (bw + vp_wid - 1, i, bw + vp_wid - 1, bw);
  vga_setcolor (LIGHTGRAY);
  vga_drawline (vp_x, bw + 1, vp_x + vp_wid, bw + 1);

  vga_setegacolor (BORDER_COLOR);
  for (i = 0; i < bw; i++) {
    vga_drawline (0, i, xmax, i);
    vga_drawline (0, i + ymax - bw, xmax, i + ymax - bw);
    vga_drawline (i, 0, i, ymax);
    vga_drawline (i + bw + vp_wid, 0, i + bw + vp_wid, ymax);
  }

#define ticks (bw+10)

  vga_setcolor (LIGHTGRAY);
  vga_drawline (0, 0, xmax, 0);
  vga_drawline (xmax, 0, xmax, ymax);
  vga_drawline (bw, bw, bw, ymax - bw);
  vga_drawline (bw, ymax - bw, xmax - bw, ymax - bw);

  vga_drawline (0, ticks, bw, ticks);
  vga_drawline (xmax - bw, ticks, xmax, ticks);
  vga_drawline (0, ymax - ticks, bw, ymax - ticks);
  vga_drawline (xmax - bw, ymax - ticks, xmax, ymax - ticks);
  vga_drawline (ticks, 0, ticks, bw);
  vga_drawline (xmax - ticks, 0, xmax - ticks, bw);
  vga_drawline (ticks, ymax - bw, ticks, ymax);
  vga_drawline (xmax - ticks, ymax - bw, xmax - ticks, ymax);

  vga_setcolor (DARKGRAY);
  vga_drawline (xmax, ymax, 0, ymax);
  vga_drawline (0, ymax, 0, 0);
  vga_drawline (bw, bw, xmax - bw, bw);
  vga_drawline (xmax - bw, bw, xmax - bw, ymax - bw);

  vga_drawline (0, ticks - 1, bw, ticks - 1);
  vga_drawline (xmax - bw, ticks - 1, xmax, ticks - 1);
  vga_drawline (0, ymax - ticks - 1, bw, ymax - ticks - 1);
  vga_drawline (xmax - bw, ymax - ticks - 1, xmax, ymax - ticks - 1);
  vga_drawline (ticks - 1, 0, ticks - 1, bw);
  vga_drawline (xmax - ticks - 1, 0, xmax - ticks - 1, bw);
  vga_drawline (ticks - 1, ymax - bw, ticks - 1, ymax);
  vga_drawline (xmax - ticks - 1, ymax - bw, xmax - ticks - 1, ymax);

  update_title (tfont, xmax >> 1);
  vga_display (vga_x, vga_y);

  while (1) {
    switch (key = getchar ()) {

    case '+':
      if (vga_key_callback && vga_key_callback(key)) {
	decode(curfile);
	force_vga_display (vga_x = 0, vga_y = 0);
      } else
	beep ();	
      break;
    case '-':
      if (vga_key_callback && vga_key_callback(key)) {
	decode(curfile);
	force_vga_display (vga_x = 0, vga_y = 0);
      } else
	beep ();	
      break;
    case 'q':
    case 'Q':
      goto end;
      break;
    case CTRL_F:
      {
	short pwid = 200,
	  pht = 32;
	char *str;

	if ((str = vga_prompt ((xmax + 1 - pwid) >> 1,
			       (ymax + 1 - pht) >> 1,
			       pwid,
			       pht,
			       "Load file: ",
			       pfont,
			       ifont,
			       YELLOW,
			       LIGHTRED,
			       BLACK,
			       PROMPT_SCROLLABLE)) == NULL ||
	    access (str, F_OK | R_OK)) {
	  beep ();
	  break;
	}
	for (i = 0; i < image.ht; i++)
	  if (image.data[i])
	    free (image.data[i]);
	if (newname)
	  free (newname);

	vga_message ((xmax + 1 - 200) >> 1,
		     (ymax + 1 - 32) >> 1,
		     200,
		     32,
		     "Decoding...",
		     pfont,
		     WHITE,
		     RED,
		     0);	/* Return immediately - no bg restoration */

	decode (curfile = newname = str);
	update_title (tfont, xmax >> 1);
	force_vga_display (vga_x = 0, vga_y = 0);
      }
      break;
    case CTRL_P:
      {
	char *str,
	 *p,
	 *f;
	short pwid = 100,
	  pht = 32;

	if ((str = vga_prompt ((xmax + 1 - pwid) >> 1,
			       (ymax + 1 - pht) >> 1,
			       pwid,
			       pht,
			       "Page: ",
			       pfont,
			       ifont,
			       YELLOW,
			       LIGHTRED,
			       BLACK,
			       PROMPT_SCROLLABLE)) == NULL) {
	  beep ();
	  break;
	}
	if (rindex (curfile, '.')) {
	  p = strdup (curfile);
	  *rindex (p, '.') = '\0';
	  f = malloc (PATH_MAX);
	  sprintf (f, NAME_TEMPLATE, p, atoi (str));
	  free (p);
	  if (!access (f, F_OK | R_OK)) {
	    for (i = 0; i < image.ht; i++)
	      if (image.data[i])
		free (image.data[i]);
	    if (newname)
	      free (newname);
	    vga_message ((xmax + 1 - 200) >> 1,
			 (ymax + 1 - 32) >> 1,
			 200,
			 32,
			 "Decoding...",
			 pfont,
			 WHITE,
			 RED,
			 0);	/* Return immediately - no time delay */

	    decode (curfile = newname = f);
	    update_title (tfont, xmax >> 1);
	    force_vga_display (vga_x = 0, vga_y = 0);
	  }
	} else {
	  if (f)
	    free (f);
	  beep ();
	}
	free (str);
      }
      break;
    case ' ':
    case 'n':
    case 'N':
      if (fname[curopt + 1] == NULL) {
	beep ();
	break;
      }
      for (i = 0; i < image.ht; i++)
	if (image.data[i])
	  free (image.data[i]);
      decode (curfile = fname[++curopt]);
      update_title (tfont, xmax >> 1);
      force_vga_display (vga_x = 0, vga_y = 0);
      break;
    case 'p':
    case 'P':
      if (curopt == optno || fname[curopt - 1] == NULL) {
	beep ();
	break;
      }
      for (i = 0; i < image.ht; i++)
	if (image.data[i])
	  free (image.data[i]);
      decode (curfile = fname[--curopt]);
      update_title (tfont, xmax >> 1);
      force_vga_display (vga_x = 0, vga_y = 0);
      break;
    case 27:
      {
	char extkey = getchar ();

	if (extkey == 91 || extkey == 79) {
	  switch (getchar ()) {
	  case ARROW_LEFT:
	    vga_x = vga_x - arrow_step < 0 ?
	      0 : vga_x - arrow_step;
	    vga_display (vga_x, vga_y);
	    break;
	  case ARROW_RIGHT:
	    vga_x = vga_x + arrow_step;
	    if (vga_x + xmax - 2 * bw > image.wid)
	      vga_x = image.wid - xmax - 2 * bw;
	    vga_display (vga_x, vga_y);
	    break;
	  case ARROW_UP:
	    vga_y = vga_y - arrow_step < 0 ?
	      0 : vga_y - arrow_step;
	    vga_display (vga_x, vga_y);
	    break;
	  case ARROW_DOWN:
	    vga_y = vga_y + arrow_step > image.ht ?
	      vga_y : vga_y + arrow_step;
	    if (vga_y + ymax - 2 * bw > image.ht)
	      vga_y = image.ht - ymax - 2 * bw;
	    vga_display (vga_x, vga_y);
	    break;
	  case PAGE_UP:
	    getchar ();
	    vga_y = vga_y - page_step < 0 ?
	      0 : vga_y - page_step;
	    vga_display (vga_x, vga_y);
	    break;
	  case PAGE_DOWN:
	    getchar ();
	    vga_y = vga_y + page_step > image.ht ?
	      vga_y : vga_y + page_step;
	    if (vga_y + ymax - 2 * bw > image.ht)
	      vga_y = image.ht - ymax - 2 * bw;
	    vga_display (vga_x, vga_y);
	    break;
	  case HOME:
	    getchar ();
	    vga_display (vga_x, vga_y = 0);
	    break;
	  case END:
	    getchar ();
	    vga_display (vga_x, vga_y = image.ht - ymax - 2 * bw);
	    break;
	  }
	}
      }
    }
  }
end:
  vga_setpalvec(0,256,rgbsave);
  vga_setmode (TEXT);
}
